# Test gradient computations.
#
# Added for pqR, 2019 Radford M. Neal.


#gctorture()  # Can enable for testing, but slow


# Check consistency of simple math derivatives using track gradient with 'D'

x <- 0.32739

print (identical (track gradient (x) gradient_of (sqrt(x)),
                  eval (D (quote (sqrt(x)), "x")) ))

print (identical (track gradient (x) gradient_of (exp(x)),
                  eval (D (quote (exp(x)), "x")) ))

print (identical (track gradient (x) gradient_of (expm1(x)),
                  eval (D (quote (expm1(x)), "x")) ))

print (identical (track gradient (x) gradient_of (log1p(x)),
                  eval (D (quote (log1p(x)), "x")) ))

print (identical (track gradient (x) gradient_of (log(x)),
                  eval (D (quote (log(x)), "x")) ))

print (identical (track gradient (x) gradient_of (cos(x)),
                  eval (D (quote (cos(x)), "x")) ))

print (identical (track gradient (x) gradient_of (sin(x)),
                  eval (D (quote (sin(x)), "x")) ))

print (identical (track gradient (x) gradient_of (tan(x)),
                  eval (D (quote (tan(x)), "x")) ))

print (identical (track gradient (x) gradient_of (acos(x)),
                  eval (D (quote (acos(x)), "x")) ))

print (identical (track gradient (x) gradient_of (asin(x)),
                  eval (D (quote (asin(x)), "x")) ))

print (identical (track gradient (x) gradient_of (cosh(x)),
                  eval (D (quote (cosh(x)), "x")) ))

print (identical (track gradient (x) gradient_of (sinh(x)),
                  eval (D (quote (sinh(x)), "x")) ))

print (identical (track gradient (x) gradient_of (tanh(x)),
                  eval (D (quote (tanh(x)), "x")) ))

print (identical (track gradient (x) gradient_of (lgamma(x)),
                  eval (D (quote (lgamma(x)), "x")) ))

print (identical (track gradient (x) gradient_of (gamma(x)),
                  eval (D (quote (gamma(x)), "x")) ))

print (identical (track gradient (x) gradient_of (digamma(x)),
                  eval (D (quote (digamma(x)), "x")) ))

print (identical (track gradient (x) gradient_of (trigamma(x)),
                  eval (D (quote (trigamma (x)), "x")) ))

print (identical (track gradient (x) gradient_of (+x),
                  eval (D (quote (+x), "x")) ))

print (identical (track gradient (x) gradient_of (-x),
                  eval (D (quote (-x), "x")) ))

print (identical (track gradient (x) gradient_of (3.1+x),
                  eval (D (quote (3.1+x), "x")) ))

print (identical (track gradient (x) gradient_of (3.1-x),
                  eval (D (quote (3.1-x), "x")) ))

print (identical (track gradient (x) gradient_of (3.1*x),
                  eval (D (quote (3.1*x), "x")) ))

print (identical (track gradient (x) gradient_of (3.1/x),
                  eval (D (quote (3.1/x), "x")) ))

print (identical (track gradient (x) gradient_of (3.1^x),
                  eval (D (quote (3.1^x), "x")) ))

print (identical (track gradient (x) gradient_of (x+3.1),
                  eval (D (quote (x+3.1), "x")) ))

print (identical (track gradient (x) gradient_of (x-3.1),
                  eval (D (quote (x-3.1), "x")) ))

print (identical (track gradient (x) gradient_of (x*3.1),
                  eval (D (quote (x*3.1), "x")) ))

print (identical (track gradient (x) gradient_of (x/3.1),
                  eval (D (quote (x/3.1), "x")) ))

print (identical (track gradient (x) gradient_of (x^3.1),
                  eval (D (quote (x^3.1), "x")) ))

print (identical (track gradient (x) gradient_of (x+x),
                  eval (D (quote (x+x), "x")) ))

print (identical (track gradient (x) gradient_of (x-x),
                  eval (D (quote (x-x), "x")) ))

print (identical (track gradient (x) gradient_of (x*x),
                  eval (D (quote (x*x), "x")) ))

print (identical (track gradient (x) gradient_of (x/x),
                  eval (D (quote (x/x), "x")) ))

print (identical (round (track gradient (x) gradient_of (x^x), 14),
                  round (eval (D (quote (x^x), "x")), 14) ))


# Check gradient computations passing through mathematical function dispatch.

a <- 256; class(a) <- "fred"
cos.fred <- function (x) tan(x)
log2.fred <- function (x) tan(tan(x))
Math.fred <- function (x,...) {
    if (.Generic=="log") sqrt(x)
    else if (.Generic=="log10") sqrt(sqrt(x))
    else if (.Generic=="exp") sqrt(sqrt(sqrt(x)))
    else NextMethod()
}

print(cos(a))
print(log2(a))
print(log(a))
print(log10(a))
print(exp(a))
print(sin(a))

print (with gradient (a) cos(a))
print (with gradient (a) log2(a))
print (with gradient (a) log(a))
print (with gradient (a) log10(a))
print (with gradient (a) exp(a))
print (with gradient (a) sin(a))


# Check tracking of gradients through S3 methods.

fuddle <- function (x,y) UseMethod("fuddle")
fuddle.default <- function (x,y) x^2+y^3
fuddle.mary <- function (x,y) sin(x) + log(y)
fuddle.bert <- function (x,y) NextMethod("fuddle")

a <- 256; class(a) <- "mary"
b <- 200; class(b) <- "bert"

with gradient (a=256,b=200) fuddle(a,b)
with gradient (a,b) fuddle(a,b)
with gradient (a,b) fuddle(b,a)

biffle <- function (x) UseMethod("biffle")
biffle.mary <- function (x) NextMethod("biffle",x,x^2,x^3)
biffle.bert <- function (x,y,z) sin(x)+cos(y)+exp(-sqrt(z)/2000)

b <- 200; class(b) <- c("mary","bert")
biffle(b)
with gradient (b) biffle(b)

`[[.xyzzy` <- 
  function (x,i) if (i==0) x$x else x$y
`[[<-.xyzzy` <- 
  function (x,i,value) { if (i==0) x$x<-value else x$y<-value; x }

with gradient (a=7) { L <- list(x=1,y=2); class(L)<-"xyzzy"; L$x<-a^2; L[[0]] }
with gradient (a=7) { L <- list(x=1,y=2); class(L)<-"xyzzy"; L[[0]]<-a^2; L }

rm(`[[.xyzzy`)
rm(`[[<-.xyzzy`)

`$.xyzzy` <- 
  function (x,i) if (i=="x") x[[1]] else x[[2]]
`$<-.xyzzy` <- 
  function (x,i,value) { if (i=="x") x[[1]]<-value else x[[2]]<-value; x }

with gradient (a=7) { L <- list(x=1,y=2); class(L)<-"xyzzy"; L[[1]]<-a^2; L$x }
with gradient (a=7) { L <- list(x=1,y=2); class(L)<-"xyzzy"; L$x<-a^2; L }

rm(`$.xyzzy`)
rm(`$<-.xyzzy`)


# Check tracking of gradients through S4 methods.  Not currently implemented,
# so this is disabled.

if (FALSE) {

setGeneric ("fiddler", function (x,y) x^2+y^3)
fiddler(2,3)
with gradient (x=2,y=3) fiddler(x,y)

}


# Check consistency of results between with gradient and numericDeriv.

x <- 0.32739
x1 <- 0.47718; x2 <- 0.89472; x3 <- 0.67325
y1 <- -0.3721; y2 <- -0.8131; y3 <- 1.22213
z1 <- 11.4319; z2 <- 13.1133; z3 <- 6.68901
w1 <- 0.8389; w2 <- 0.1123; w3 <- 4.68701
v1 <- 3; v2 <- 7; v3 <- 0.6513
u1 <- -0.3721; u2 <- 1.8131; u3 <- 1.22213
i1 <- 3

bindgrads <- function (r1,r2) 
    cbind (rbind(r1,r2), rbind(attr(r1,"gradient"),unlist(attr(r2,"gradient"))))

test1 <- function (fun,...)
    print (bindgrads (numericDeriv(quote(fun(x,...)),"x"),
                      with gradient (x) fun(x,...)))
test1p1 <- function (fun,...)
    print (bindgrads (numericDeriv(quote(fun(x+1,...)),"x"),
                      with gradient (x) fun(x+1,...)))

test1r <- function (fun,...) {
    f <- function (x) { set.seed(179); fun(1,x,...) }
    print (bindgrads (numericDeriv(quote(f(x)),"x"),
                      with gradient (x) f(x)))
}

test2 <- function (fun,...) {
    print (bindgrads (numericDeriv(quote(fun(x1,x2,...)),"x1"),
                      with gradient (x1) fun(x1,x2,...)))
    print (bindgrads (numericDeriv(quote(fun(x1,x2,...)),"x2"),
                      with gradient (x2) fun(x1,x2,...)))
    print (bindgrads (numericDeriv(quote(fun(x1,x2,...)),c("x1","x2")),
                      with gradient (x1,x2) fun(x1,x2,...)))
    print (bindgrads (numericDeriv(quote(fun(x1,x2,...)),c("x1","x2")),
      { r <- with gradient (x1) { s <- with gradient (x2) fun(x1,x2,...); 
                                  g2 <<- attr(s,"gradient"); s }
        attr(r,"gradient") <- cbind(g1=attr(r,"gradient"),g2=g2)
        r
      }
    ))
}

test2y <- function (fun,...) {
    print (bindgrads (numericDeriv(quote(fun(y1,y3,...)),"y1"),
                      with gradient (y1) fun(y1,y3,...)))
    print (bindgrads (numericDeriv(quote(fun(y1,y3,...)),"y3"),
                      with gradient (y3) fun(y1,y3,...)))
    print (bindgrads (numericDeriv(quote(fun(y1,y3,...)),c("y1","y3")),
                      with gradient (y1,y3) fun(y1,y3,...)))
    print (bindgrads (numericDeriv(quote(fun(y1,y3,...)),c("y1","y3")),
      { r <- with gradient (y1) { s <- with gradient (y3) fun(y1,y3,...); 
                                  g2 <<- attr(s,"gradient"); s }
        attr(r,"gradient") <- cbind(g1=attr(r,"gradient"),g2=g2)
        r
      }
    ))
}

test2z <- function (fun,...) {
    print (bindgrads (numericDeriv(quote(fun(z1,z2,...)),"z1"),
                      with gradient (z1) fun(z1,z2,...)))
    print (bindgrads (numericDeriv(quote(fun(z1,z2,...)),"z2"),
                      with gradient (z2) fun(z1,z2,...)))
    print (bindgrads (numericDeriv(quote(fun(z1,z2,...)),c("z1","z2")),
                      with gradient (z1,z2) fun(z1,z2,...)))
    print (bindgrads (numericDeriv(quote(fun(z1,z2,...)),c("z1","z2")),
      { r <- with gradient (z1) { s <- with gradient (z2) fun(z1,z2,...); 
                                  g2 <<- attr(s,"gradient"); s }
        attr(r,"gradient") <- cbind(g1=attr(r,"gradient"),g2=g2)
        r
      }
    ))
}

test2u <- function (fun,...) {
    print (bindgrads (numericDeriv(quote(fun(u1,u2,...)),c("u1","u2")),
                      with gradient (u1,u2) fun(u1,u2,...)))
}

test2i <- function (fun,...) {
    print (bindgrads (numericDeriv(quote(fun(i1,x2,...)),"x2"),
                      with gradient (x2) fun(i1,x2,...)))
}

test2r <- function (fun,...) {
    f <- function (x1,x2) { set.seed(179); fun(1,x1,x2,...) }
    print (bindgrads (numericDeriv(quote(f(x1,x2)),c("x1","x2")),
                      with gradient (x1,x2) f(x1,x2)))
}

test3 <- function (fun,...) {
    print (bindgrads (numericDeriv(quote(fun(x1,x2,x3,...)),"x1"),
                      with gradient (x1) fun(x1,x2,x3,...)))
    print (bindgrads (numericDeriv(quote(fun(x1,x2,x3,...)),"x2"),
                      with gradient (x2) fun(x1,x2,x3,...)))
    print (bindgrads (numericDeriv(quote(fun(x1,x2,x3,...)),"x3"),
                      with gradient (x3) fun(x1,x2,x3,...)))
    print (bindgrads (numericDeriv(quote(fun(x1,x2,x3,...)),c("x1","x2","x3")),
                      with gradient (x1,x2,x3) fun(x1,x2,x3,...)))
}

test3y <- function (fun,...) {
    print (bindgrads (numericDeriv(quote(fun(y1,y2,y3,...)),"y1"),
                      with gradient (y1) fun(y1,y2,y3,...)))
    print (bindgrads (numericDeriv(quote(fun(y1,y2,y3,...)),"y2"),
                      with gradient (y2) fun(y1,y2,y3,...)))
    print (bindgrads (numericDeriv(quote(fun(y1,y2,y3,...)),"y3"),
                      with gradient (y3) fun(y1,y2,y3,...)))
    print (bindgrads (numericDeriv(quote(fun(y1,y2,y3,...)),c("y1","y2","y3")),
                      with gradient (y1,y2,y3) fun(y1,y2,y3,...)))
}

test3z <- function (fun,...) {
    print (bindgrads (numericDeriv(quote(fun(z1,z2,z3,...)),"z1"),
                      with gradient (z1) fun(z1,z2,z3,...)))
    print (bindgrads (numericDeriv(quote(fun(z1,z2,z3,...)),"z2"),
                      with gradient (z2) fun(z1,z2,z3,...)))
    print (bindgrads (numericDeriv(quote(fun(z1,z2,z3,...)),"z3"),
                      with gradient (z3) fun(z1,z2,z3,...)))
    print (bindgrads (numericDeriv(quote(fun(z1,z2,z3,...)),c("z1","z2","z3")),
                      with gradient (z1,z2,z3) fun(z1,z2,z3,...)))
}

test3w <- function (fun,...) {
    print (bindgrads (numericDeriv(quote(fun(w1,w2,w3,...)),"w1"),
                      with gradient (w1) fun(w1,w2,w3,...)))
    print (bindgrads (numericDeriv(quote(fun(w1,w2,w3,...)),"w2"),
                      with gradient (w2) fun(w1,w2,w3,...)))
    print (bindgrads (numericDeriv(quote(fun(w1,w2,w3,...)),"w3"),
                      with gradient (w3) fun(w1,w2,w3,...)))
    print (bindgrads (numericDeriv(quote(fun(w1,w2,w3,...)),c("w1","w2","w3")),
                      with gradient (w1,w2,w3) fun(w1,w2,w3,...)))
}

test3u <- function (fun,...) {
    print (bindgrads (numericDeriv(quote(fun(u1,u2,u3,...)),"u1"),
                      with gradient (u1) fun(u1,u2,u3,...)))
    print (bindgrads (numericDeriv(quote(fun(u1,u2,u3,...)),"u2"),
                      with gradient (u2) fun(u1,u2,u3,...)))
    print (bindgrads (numericDeriv(quote(fun(u1,u2,u3,...)),"u3"),
                      with gradient (u3) fun(u1,u2,u3,...)))
    print (bindgrads (numericDeriv(quote(fun(u1,u2,u3,...)),c("u1","u2","u3")),
                      with gradient (u1,u2,u3) fun(u1,u2,u3,...)))
}

test3v <- function (fun,...) {
    print (bindgrads (numericDeriv(quote(fun(v1,v2,v3,...)),"v3"),
                      with gradient (v3) fun(v1,v2,v3,...)))
}

test1(abs)

test1(sqrt)

test1(exp)
test1(expm1)

test1(log1p)
test1(log)
test1(log2)
test1(log10)

test1(cos)
test1(sin)
test1(tan)

test1(acos)
test1(asin)
test1(atan)

test1(cosh)
test1(sinh)
test1(tanh)

test1p1(acosh)
test1(asinh)
test1(atanh)

test1(gamma)
test1(lgamma)
test1(digamma)
test1(trigamma)
test1(psigamma)
test1p1(psigamma)
test1(psigamma,3)
test1p1(psigamma,3)

test2(atan2)
test2z(atan2)

test2(beta)
test2z(beta)

test2(lbeta)
test2z(lbeta)

test2(dchisq)
test2(dchisq,log=TRUE)

test2(pchisq)
test2(pchisq,log=TRUE)
test2(pchisq,lower=FALSE)
test2(pchisq,log=TRUE,lower=FALSE)

test2(qchisq)
test2u(qchisq,log=TRUE)
test2(qchisq,lower=FALSE)
test2u(qchisq,log=TRUE,lower=FALSE)

test2(dexp)
test2(dexp,log=TRUE)

test2(pexp)
test2(pexp,log=TRUE)
test2(pexp,lower=FALSE)
test2(pexp,log=TRUE,lower=FALSE)

test2(qexp)
test2y(qexp,log=TRUE)
test2(qexp,lower=FALSE)
test2y(qexp,log=TRUE,lower=FALSE)

test1r(rexp)

test2i(dgeom)
test2i(dgeom,log=TRUE)

test2i(pgeom)
test2i(pgeom,log=TRUE)
test2i(pgeom,lower=FALSE)
test2i(pgeom,log=TRUE,lower=FALSE)

test2i(dpois)
test2i(dpois,log=TRUE)

test2i(ppois)
test2i(ppois,log=TRUE)
test2i(ppois,lower=FALSE)
test2i(ppois,log=TRUE,lower=FALSE)

test2(dt)
test2(dt,log=TRUE)
test2z(dt)
test2z(dt,log=TRUE)

test2(pt)
test2(pt,log=TRUE)
test2(pt,lower=FALSE)
test2(pt,log=TRUE,lower=FALSE)
test2z(pt)
test2z(pt,log=TRUE)
test2z(pt,lower=FALSE)
test2z(pt,log=TRUE,lower=FALSE)

test2(qt)
test2u(qt,log=TRUE)
test2(qt,lower=FALSE)
test2u(qt,log=TRUE,lower=FALSE)

test3(dbeta)
test3(dbeta,log=TRUE)
test3w(dbeta)
test3w(dbeta,log=TRUE)

test3(pbeta)
test3(pbeta,log=TRUE)
test3(pbeta,lower=FALSE)
test3(pbeta,log=TRUE,lower=FALSE)

test3(qbeta)
test3u(qbeta,log=TRUE)
test3(qbeta,lower=FALSE)
test3u(qbeta,log=TRUE,lower=FALSE)

test3v(dbinom)
test3v(dbinom,log=TRUE)

test3v(pbinom)
test3v(pbinom,log=TRUE)
test3v(pbinom,lower=FALSE)
test3v(pbinom,log=TRUE,lower=FALSE)

test3(dcauchy)
test3(dcauchy,log=TRUE)
test3z(dcauchy)
test3z(dcauchy,log=TRUE)

test3(pcauchy)
test3(pcauchy,log=TRUE)
test3(pcauchy,lower=FALSE)
test3(pcauchy,log=TRUE,lower=FALSE)
test3z(pcauchy)
test3z(pcauchy,log=TRUE)
test3z(pcauchy,lower=FALSE)
test3z(pcauchy,log=TRUE,lower=FALSE)

test3(qcauchy)
test3y(qcauchy,log=TRUE)
test3(qcauchy,lower=FALSE)
test3y(qcauchy,log=TRUE,lower=FALSE)

test2r(rcauchy)

test3(df)
test3(df,log=TRUE)

test3(pf)
test3(pf,log=TRUE)
test3(pf,lower=FALSE)
test3(pf,log=TRUE,lower=FALSE)

test3(qf)
test3u(qf,log=TRUE)
test3(qf,lower=FALSE)
test3u(qf,log=TRUE,lower=FALSE)

test3(dgamma)
test3(dgamma,log=TRUE)
test3z(dgamma)
test3z(dgamma,log=TRUE)

dgamma_scale <- function (x,a,b,...) dgamma(x,a,scale=b,...)
test3(dgamma_scale)
test3(dgamma_scale,log=TRUE)
test3z(dgamma_scale)
test3z(dgamma_scale,log=TRUE)

test3(pgamma)
test3(pgamma,log=TRUE)
test3(pgamma,lower=FALSE)
test3(pgamma,log=TRUE,lower=FALSE)
test3z(pgamma)
test3z(pgamma,log=TRUE)
test3z(pgamma,lower=FALSE)
test3z(pgamma,log=TRUE,lower=FALSE)

test3(qgamma)
test3u(qgamma,log=TRUE)
test3(qgamma,lower=FALSE)
test3u(qgamma,log=TRUE,lower=FALSE)

test3(dlnorm)
test3(dlnorm,log=TRUE)
test3z(dlnorm)
test3z(dlnorm,log=TRUE)

test3(plnorm)
test3(plnorm,log=TRUE)
test3(plnorm,lower=FALSE)
test3(plnorm,log=TRUE,lower=FALSE)
test3z(plnorm)
test3z(plnorm,log=TRUE)
test3z(plnorm,lower=FALSE)
test3z(plnorm,log=TRUE,lower=FALSE)

test3(qlnorm)
test3y(qlnorm,log=TRUE)
test3(qlnorm,lower=FALSE)
test3y(qlnorm,log=TRUE,lower=FALSE)

test2r(rlnorm)

test3(dlogis)
test3(dlogis,log=TRUE)
test3z(dlogis)
test3z(dlogis,log=TRUE)

test3(plogis)
test3(plogis,log=TRUE)
test3(plogis,lower=FALSE)
test3(plogis,log=TRUE,lower=FALSE)
test3z(plogis)
test3z(plogis,log=TRUE)
test3z(plogis,lower=FALSE)
test3z(plogis,log=TRUE,lower=FALSE)

test3(qlogis)
test3y(qlogis,log=TRUE)
test3(qlogis,lower=FALSE)
test3y(qlogis,log=TRUE,lower=FALSE)

test2r(rlogis)

test3v(dnbinom)
test3v(dnbinom,log=TRUE)

test3v(pnbinom)
test3v(pnbinom,log=TRUE)
test3v(pnbinom,lower=FALSE)
test3v(pnbinom,log=TRUE,lower=FALSE)

dnbinom_mu <- function (x, size, mu, ...) dnbinom(x,size,mu=mu,...)
test3v(dnbinom_mu)
test3v(dnbinom_mu,log=TRUE)

pnbinom_mu <- function (q, size, mu, ...) pnbinom(q,size,mu=mu,...)
test3v(pnbinom_mu)
test3v(pnbinom_mu,log=TRUE)
test3v(pnbinom_mu,lower=FALSE)
test3v(pnbinom_mu,log=TRUE,lower=FALSE)

test3(dnorm)
test3(dnorm,log=TRUE)
test3z(dnorm)
test3z(dnorm,log=TRUE)

test3(pnorm)
test3(pnorm,log=TRUE)
test3(pnorm,lower=FALSE)
test3(pnorm,log=TRUE,lower=FALSE)
test3z(pnorm)
test3z(pnorm,log=TRUE)
test3z(pnorm,lower=FALSE)
test3z(pnorm,log=TRUE,lower=FALSE)

test3(qnorm)
test3y(qnorm,log=TRUE)
test3(qnorm,lower=FALSE)
test3y(qnorm,log=TRUE,lower=FALSE)

test2r(rnorm)

test3w(dunif)
test3w(dunif,log=TRUE)

test3w(punif)
test3w(punif,log=TRUE)
test3w(punif,lower=FALSE)
test3w(punif,log=TRUE,lower=FALSE)

test3w(qunif)
test3y(qunif,log=TRUE)
test3w(qunif,lower=FALSE)
test3y(qunif,log=TRUE,lower=FALSE)

test2r(runif)

test3(dweibull)

test3(pweibull)
test3(pweibull,log=TRUE)
test3(pweibull,lower=FALSE)
test3(pweibull,log=TRUE,lower=FALSE)

test3(qweibull)
test3u(qweibull,log=TRUE)
test3(qweibull,lower=FALSE)
test3u(qweibull,log=TRUE,lower=FALSE)

test2r(rweibull)


# Check gradients of lists.

with gradient (a=3.1) list (a^2, a^3)
with gradient (a=3.1) list (list(a^2), xx=sin(a), yy=99, zz=list(7,a^3))
with gradient (a=3.1) with gradient (b=7.7) list(a*b,a^2*b^2)

with gradient (a=3.1) with gradient (b=100*a) list(a, b, a^2, b^2, a*b, a^2*b^2)

with gradient (a=3.1,b=4.2) list(a,b,a*b)

with gradient (a=3.1,b=4.2) with gradient (c=7.9) list (a*b*c)
with gradient (c=7.9) with gradient (a=3.1,b=4.2) list (a*b*c)

with gradient (a=3.1,b=4.2) with gradient (c=7.9*a) list (a*b*c)
with gradient (c=7.9) with gradient (a=3.1*c,b=4.2) list (a*b*c)

with gradient (a=3.1) { x <- list(xxx=a,yyy=a^2); list(qqq=x,rrr=a^3) }

with gradient (a=3.1, b=4.2) {
  r <- list(abc=a^2,pqr=b^2,xyz=a*b)
  print(gradient_of(r$pqr))
  print(gradient_of(r$abc))
  print(gradient_of(r$xyz))
  r
}


# Test gradients of list elements found with $ and [[.]].

with gradient (a=9) { r <- list(x=a,y=a^2,z=a^3); list(r$z,r$y) }
with gradient (a=9) { r <- list(x=a,y=a^2,z=a^3); list(r[[3]],r[[2]]) }

with gradient (a=9) { r <- list(x=a,y=list(p=a^2,7,q=a^3))
                      list(r$y$q,r$x) }
with gradient (a=9) { r <- list(x=a,y=list(p=a^2,7,q=a^3))
                      list(r[[2]][[3]],r[[1]]) }
with gradient (a=9) { r <- list(x=a,y=list(p=a^2,7,q=a^3))
                      list(r[[c(2,3)]],r[[1]]) }


# Test gradients of sub-lists found with [.].

with gradient (a=9) { r <- list(x=a,y=a^2,z=a^3); r[2] }
with gradient (a=9) { r <- list(x=a,y=a^2,z=a^3); r[c(2,1)] }
with gradient (a=9) { r <- list(x=a,y=a^2,z=a^3); r[1..2] }
with gradient (a=9) { r <- list(x=a,y=a^2,z=a^3); r[c(2,1,1,3)] }
with gradient (a=9) { r <- list(x=a,y=a^2,z=a^3); r[c(5,1)] }
with gradient (a=9) { r <- list(x=a,y=a^2,z=a^3); r[1..5] }

L <- matrix (list(99,33,44,66,22,11),3,2)
G <- with gradient (L) L[3,c(2,2,1)]
L[,]
G[]
str(G)

L <- matrix (list(99,33,44,66,22,11),3,2)
G <- with gradient (L) L[1,c(2,2,1)]
L[,]
G[]
str(G)

L <- matrix (list(99,33,44,66,22,11),3,2)
G <- with gradient (L) L[2..3,c(2,1,2)]
L[,]
G[,]
str(G)

L <- matrix (list(99,33,44,66,22,11),3,2)
G <- with gradient (L) L[c(2,3),c(2,1,2)]
L[,]
G[,]
str(G)

L <- matrix (list(99,33,44,66,22,11),3,2)
G <- with gradient (L) { L[[2,2]] <- L[[2,2]]^2; L[2..3,c(2,1,2)] }
L[,]
G[,]
str(G)

L <- array (list(9,88,7,33,44,999,20,303,111,1111,222,3333,5555,330,1,0,-3,-2),
            c(3,2,3))
G <- with gradient (L) L[c(3,1,1),c(2,2,1),c(2,1,3)]
L[,,]
G[,,]
str(G)


# Test gradients of vector elements found with [[.]].

with gradient (a=9) {
  r <- numeric(4); r[[1]] <- a; r[[2]] <- a^2; r[[3]] <- a^3; r[[4]] <- sin(a)
  list(r[[3]],r[[2]])
}

with gradient (a=c(9,5,2)) {
  r <- numeric(4)
  r[[1]] <- a[[2]]
  r[[2]] <- a[[3]]^2 + a[[1]]
  r[[3]] <- a[[3]]^3
  r[[4]] <- sin(a[[1]])
  list(r[[3]],r[[2]],r[[4]],r[[1]])
}

with gradient (a=c(9,5,2)) {
  r <- numeric(4)
  dim(r) <- c(2,2)
  r[[1]] <- a[[2]]
  r[[2]] <- a[[3]]^2 + a[[1]]
  r[[3]] <- a[[3]]^3
  r[[4]] <- sin(a[[1]])
  list(r[[2,1]],r[[1,2]],r[[1,1]],r[[2,2]])
}

with gradient (a=c(9,5,2)) {
  r <- numeric(12)
  dim(r) <- c(2,3,2)
  r[[1]] <- a[[2]]
  r[[2]] <- a[[3]]^2 + a[[1]]
  r[[3]] <- a[[3]]^3
  r[[12]] <- sin(a[[1]])
  list(r[[1,1,1]],r[[2,1,1]],r[[2,3,2]])
}


# Test gradients of sub-vectors found with [.].

with gradient (a=9) {
  r <- numeric(4); r[[1]] <- a; r[[2]] <- a^2; r[[3]] <- a^3; r[[4]] <- sin(a)
  r[c(2,1,1,4)]
}

with gradient (a=9) {
  r <- numeric(4); r[[1]] <- a; r[[2]] <- a^2; r[[3]] <- a^3; r[[4]] <- sin(a)
  r[2..3]
}

with gradient (a=c(9,5,2)) {
  r <- numeric(4)
  r[[1]] <- a[[2]]
  r[[2]] <- a[[3]]^2 + a[[1]]
  r[[3]] <- a[[3]]^3
  r[[4]] <- sin(a[[1]])
  r[c(2,1,1,4)]
}

with gradient (a=c(9,5,2)) {
  r <- numeric(4)
  r[[1]] <- a[[2]]
  r[[2]] <- a[[3]]^2 + a[[1]]
  r[[3]] <- a[[3]]^3
  r[[4]] <- sin(a[[1]])
  r[2..3]
}

with gradient (a=c(9,5,2)) {
  r <- numeric(4)
  dim(r) <- c(2,2)
  r[[1]] <- a[[2]]
  r[[2]] <- a[[3]]^2 + a[[1]]
  r[[3]] <- a[[3]]^3
  r[[4]] <- sin(a[[1]])
  list(r[2,1],r[1,2],r[1,1],r[2,2])
}

with gradient (a=c(9,5,2)) {
  r <- numeric(6)
  dim(r) <- c(3,2)
  r[[1]] <- a[[2]]
  r[[2]] <- a[[3]]^2 + a[[1]]
  r[[3]] <- a[[3]]^3
  r[[4]] <- sin(a[[1]])
  list(r[2..3,1],r[1..2,c(2,1)])
}

with gradient (a=c(9,5,2)) {
  r <- numeric(6)
  dim(r) <- c(3,2)
  r[[1]] <- a[[2]]
  r[[2]] <- a[[3]]^2 + a[[1]]
  r[[3]] <- a[[3]]^3
  r[[4]] <- sin(a[[1]])
  list(r[c(2,3),1],r[c(1,2),c(2,1)])
}

with gradient (a=c(9,5,2)) {
  r <- numeric(4)
  dim(r) <- c(2,2)
  r[[1]] <- a[[2]]
  r[[2]] <- a[[3]]^2 + a[[1]]
  r[[3]] <- a[[3]]^3
  r[[4]] <- sin(a[[1]])
  list(r[2,c(1,1,2)],r[2,c(1,1,2)]^2)
}

with gradient (a=c(9,5,2)) {
  r <- numeric(12)
  dim(r) <- c(2,3,2)
  r[[1]] <- a[[2]]
  r[[2]] <- a[[3]]^2 + a[[1]]
  r[[3]] <- a[[3]]^3
  r[[12]] <- sin(a[[1]])
  list(r[1,1,1],r[2,1,1],r[2,3,2])
}

with gradient (a=c(9,5,2)) {
  r <- numeric(12)
  dim(r) <- c(2,3,2)
  r[[1]] <- a[[2]]
  r[[2]] <- a[[3]]^2 + a[[1]]
  r[[3]] <- a[[3]]^3
  r[[8]] <- a[[3]]
  r[[10]] <- sqrt(a[[1]])
  r[[12]] <- sin(a[[1]])
  list(r[c(2,1),1,1],r[2,2,2],r[2,c(3,1,1,2),2])
}

with gradient (a=c(3,5)) a[1..5]
with gradient (a=c(3,5)) a[c(5,1,7,2,3)]
with gradient (a=list(3,5)) a[1..5]
with gradient (a=list(3,5)) a[c(5,1,7,2,3)]


# Test 'compute gradient'.

f <- function (x) compute gradient (x) x^2 as 123*x # wrong, see if happens

with gradient (a=12) f(a)
with gradient (a=12) f(a^2)
with gradient (a=c(1,12)) f(a^2)

g <- function (x,y)  # assume length(x) == length(y)
  compute gradient (x,y) list(x+y,x-y,x*y)
  as # wrong, but see if it happens
     list (diag(x*y,length(x)), diag(x+y,length(x)), diag(x+y,length(x))), 
     list (diag(x-y,length(x)), diag(x-y,length(x)), diag(7*x*y,length(x)))

with gradient (z=9) g(100*z,z+1)
with gradient (z=c(9,7)) g(100*z,z+1)

g <- function (x,y)  # assume length(x) == length(y)
  compute gradient (x,y) list(x+y,x-y,x*y)
  as # wrong, but see if it happens
     list (x*y, x+y, x+y), 
     list (x-y, x-y, 7*x*y)

with gradient (z=9) g(100*z,z+1)
with gradient (z=c(9,7)) g(100*z,z+1)

g <- function (x,y)  # assume length(x) == length(y)
  compute gradient (x,y) list(x+y,x-y,x*y)
  as # wrong, but see if it happens
     list (function () x*y, function () x+y, function () x+y),
     list (function () x-y, function () x-y, function () 7*x*y)

with gradient (z=9) g(100*z,z+1)
with gradient (z=c(9,7)) g(100*z,z+1)


# Test subassign operations.

with gradient (a=3.01) {
  r <- list(x=a^2,y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  r$y <- 1/a
  r
}

with gradient (a=3.01) {
  r <- list(x=a^2,y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  r[[2]] <- 1/a
  r
}

with gradient (a=3.01) {
  r <- list(x=a^2,y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  r$y <- NULL
  r
}

with gradient (a=3.01) {
  r <- list(x=a^2,y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  r[[2]] <- NULL
  r
}

with gradient (a=3.01) {
  r <- list(x=a^2,y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  r[2] <- NULL
  r
}

with gradient (a=3.01) {
  r <- list(x=a^2,y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  r[1..2] <- NULL
  r
}

with gradient (a=3.01) {
  r <- list(x=a^2,y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  r[c(2,2,1,1,2)] <- NULL
  r
}

with gradient (a=3.01) {
  r <- list(x=a^2,y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  r$w <- -a
  r
}

with gradient (a=3.01) {
  r <- list(x=a^2,y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  r[[4]] <- -a
  r
}

with gradient (a=3.01) {
  r <- list(x=a^2,y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  r[4] <- list(-a)
  r
}

with gradient (a=3.01) {
  r <- list(x=a^2,y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  r[4] <- -a
  r
}

with gradient (a=c(6,3,5)) { a[2] <- list(a[[1]]^2); a }

with gradient (a=3.01) {
  r <- list(w=-a,x=a^2,y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  r[c(2,4)] <- list(10*a,7*a)
  r
}

with gradient (a=3.01) {
  r <- list(a=3,x=list(v=8,w=3*a,x=4*a),y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  r$x$w <- 1/a
  r
}

with gradient (a=3.01) {
  r <- list(a=3,x=list(v=8,w=3*a,x=4*a),y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  r[[2]]$w <- 1/a
  r
}

with gradient (a=3.01) {
  r <- list(a=3,x=list(v=8,w=3*a,x=4*a),y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  r$x[[2]] <- 1/a
  r
}

with gradient (a=3.01) {
  r <- list(a=3,x=list(v=8,w=3*a,x=4*a),y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  r$x[2] <- list(1/a)
  r
}

with gradient (a=3.01) {
  r <- list(a=3,x=list(v=8,w=3*a,x=4*a),y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  r[[2]][[2]] <- 1/a
  r
}

with gradient (a=3.01) {
  r <- list(a=3,x=list(v=8,w=3*a,x=4*a),y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  r[[2]][2] <- list(1/a)
  r
}

with gradient (a=3.01) {
  r <- list(a=3,x=list(v=8,w=3*a,x=4*a),y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  r$x$w <- NULL
  r
}

with gradient (a=3.01) {
  r <- list(a=3,x=list(v=8,w=3*a,x=4*a),y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  r[[2]][[2]] <- NULL
  r
}

if (FALSE)  # not supported yet
with gradient (a=3.01) {
  r <- list(a=3,x=list(v=8,w=3*a,x=4*a),y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  r[[c(2,2)]] <- NULL
  r
}

with gradient (a=3.01) {
  r <- list(a=3,x=list(v=8,w=3*a,x=4*a),y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  r$x$u <- -a
  r
}

with gradient (a=3.01) {
  r <- list(a=3,x=list(v=8,w=3*a,x=4*a),y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  r[[2]][[4]] <- -a
  r
}

with gradient (a=c(5.1,7.2,3.9,0.7)) { a[3] <- 4; a }
with gradient (a=c(5.1,7.2,3.9,0.7)) { a[c(3,1)] <- 4; a }
with gradient (a=c(5.1,7.2,3.9,0.7)) { a[c(3,1)] <- a[4]^2; a }
with gradient (a=c(5.1,7.2,3.9,0.7)) { b<-numeric(5); b[4..5]<-a[c(3,1)]; b }

with gradient (a=matrix(list(1.0,7.2,3.9,0.7,9.9,1.2,8.8,6.0),2,4)) { 
  a[1,1] <- a[2,4]
  a[2,c(1,3)] <- a[[1,1]]^2
  a[1,c(1,2)] <- list(a[[1,1]]^3,a[[1,1]]^4)
  a[1,4] <- 99
  a
}

with gradient (a=matrix(c(1.0,7.2,3.9,0.7,9.9,1.2,8.8,6.0),2,4)) { 
  a[1,1] <- a[2,4]
  a[2,c(1,3)] <- a[[1,1]]^2
  a[1,c(1,2)] <- a[[1,1]]^c(3,4)
  a[1,4] <- 99
  a
}

with gradient (a=array(list(1.0,7.2,3.9,0.7,9.9,1.2,8.8,6.0),c(2,2,2))) { 
  a[1,1,1] <- a[2,2,2]
  a[2,c(1,2),c(2,1)] <- a[[1,1,1]]^2
  a[1,c(1,2),1] <- list(a[[1,1,1]]^3,a[[1,1,1]]^4)
  a[1,2,2] <- 99
  a
}

with gradient (a=array(c(1.0,7.2,3.9,0.7,9.9,1.2,8.8,6.0),c(2,2,2))) { 
  a[1,1,1] <- a[2,2,2]
  a[2,c(1,2),c(2,1)] <- a[[1,1,1]]^2
  a[1,c(1,2),1] <- a[[1,1,1]]^c(3,4)
  a[1,2,2] <- 99
  a
}

with gradient (L=list(c(9,7,6),-2)) { L[[1]][[2]] <- 3; L }


# Subassignment with user-defined subassignment functions.

xel <- function (a) a$x
`xel<-` <- function (a,value) { a$x <- value; a }

with gradient (a=3.01) {
  r <- list(x=a^2,y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  xel(r) <- 1/a
  r
}

with gradient (a=3.01) {
  r <- list(x=a^2,y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  xel(r) <- NULL
  r
}

with gradient (a=3.01) {
  r <- list(y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  xel(r) <- -a^2
  r
}

with gradient (a=3.01) {
  r <- list(x=list(w=3*a,x=4*a),y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  xel(xel(r)) <- 1/a;
  r
}

with gradient (a=3.01) {
  r <- list(x=list(w=3*a),y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  xel(xel(r)) <- -a^2
  r
}

elip1 <- function (a,i) a[[i+1]]
`elip1<-` <- function (a,i,value) { a[[i+1]] <- value; a }

with gradient (a=3.01) {
  r <- list(x=a^2,y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  elip1(r,0) <- 1/a
  r
}

with gradient (a=3.01) {
  r <- list(x=a^2,y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  elip1(r,0) <- NULL
  r
}

with gradient (a=3.01) {
  r <- list(y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  elip1(r,2) <- -a^2
  r
}

with gradient (a=3.01) {
  r <- list(x=list(w=3*a,x=4*a),y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  elip1(elip1(r,0),1) <- 1/a;
  r
}

with gradient (a=3.01) {
  r <- list(x=list(w=3*a),y=a^3,z=a^4)
  print(gradient_of(r))
  cat("--\n")
  elip1(elip1(r,0),1) <- -a^2
  r
}

with gradient (a=3.01) {
  r <- list(x=1,y=2)
  class(r) <- "fred"
  r$y <- a^2
  r
}

with gradient (a=3.01) {
  r <- list(x=1,y=2)
  class(r) <- "fred"
  r[[2]] <- a^2
  r
}

with gradient (a=3.01) {
  r <- list(x=1,y=2)
  class(r) <- "fred"
  r$x <- -a
  r$y <- a^2
  r
}

with gradient (a=3.01) {
  r <- list(x=1,y=2)
  class(r) <- "fred"
  r$x <- -a
  r[[2]] <- a^2
  r
}

with gradient (a=3.01) {
  r <- list(x=1,y=2)
  class(r) <- "fred"
  r <- `$<-` (r, "y", a^2)
  r
}

with gradient (a=3.01) {
  r <- list(x=1,y=2)
  class(r) <- "fred"
  r <- `[[<-` (r, 2, a^2)
  r
}

with gradient (a=3.01) {
  r <- list(x=-1,y=2)
  class(r) <- "fred"
  r$x <- -a
  r <- `$<-` (r, "y", a^2)
  r
}

with gradient (a=3.01) {
  r <- list(x=-1,y=2)
  class(r) <- "fred"
  r$x <- -a
  r <- `[[<-` (r, 2, a^2)
  r
}

with gradient (a=3.01) {
  r <- list(x=-1,y=2,z=7)
  class(r) <- "fred"
  r$x <- -a
  r <- `[<-` (r, c(3,2), list(a^2,a^3))
  r
}


# Test gradients w.r.t. list values.

with gradient (a=list(5,55,555)) a

with gradient (a=list(x=5,y=55,z=555)) a$x
with gradient (a=list(x=5,y=55,z=555)) a$y
with gradient (a=list(x=5,y=55,z=555)) a$z

with gradient (a=list(x=5,y=55,z=555)) sqrt(a$y)
with gradient (a=list(x=5,y=55,z=555)) a$z+a$x
with gradient (a=list(x=5,y=55,z=555)) a$z*a$x

with gradient (a=list(x=5,y=55,z=555)) list(a$z,a$x^2, a$x*a$y*a$z)

with gradient (a=list(x=7,y=list(3,z=10))) a
with gradient (a=list(x=7,y=list(3,z=10))) list(a$y$z,a$x,-a$x)
with gradient (a=list(x=7,y=list(3,z=10))) list(a$y,a$x,-a$x)

with gradient (b=list(3,abc=4)) {
  r <- list(x=b,y=b$abc,z=b$abc^2)
  r$y <- NULL
  r
}

with gradient (b=list(3,abc=4)) {
  r <- list(x=b,y=b$abc,z=b$abc^2)
  r$x <- NULL
  r
}

with gradient (b=list(10,bb=list(bbb=3,4),20)) {
  r <- list(a1=7,a2=-b$bb$bbb,a3=b)
  r$a3 <- NULL
  r
}

with gradient (b=list(3,abc=4)) {
  r <- list(x=b,y=b$abc,z=b$abc^2)
  r$y <- b$abc^3
  r
}

with gradient (b=list(3,abc=4)) {
  r <- list(x=b,y=b$abc,z=b$abc^2)
  r$w <- b$abc^3
  r
}

fsub <- function (...) `[[`(...)
with gradient (a=list(7,8)) fsub(a,1)


# Test gradients w.r.t. numeric vectors.

with gradient (a=c(5,55,555)) a

with gradient (a=c(x=5,y=55,z=555)) a[["x"]]
with gradient (a=c(x=5,y=55,z=555)) a[["y"]]
with gradient (a=c(x=5,y=55,z=555)) a[["z"]]

with gradient (a=c(x=5,y=55,z=555)) sqrt(a[["y"]])
with gradient (a=c(x=5,y=55,z=555)) a[["z"]]+a[["x"]]
with gradient (a=c(x=5,y=55,z=555)) 2.1*a[["z"]]+3.4*a[["x"]]

with gradient (a=c(9.1,4.1)) {
  b <- numeric(3)
  b[[2]] <- a[[1]]
  b[[1]] <- 3*a[[2]] + b[[2]]
  b[[3]] <- sin(a[[2]])
  b[[5]] <- b[[1]]^2
  b
}

with gradient (a=c(5.1,2.2,3.7)) {
  x <- numeric(2)
  x[[1]] <- a[[2]]^2
  x[[2]] <- a[[3]]
  sin(x)
}

with gradient (a=c(5.1,2.2,3.7),b=c(7.3,6.4)) {
  x <- numeric(2)
  x[[1]] <- a[[2]]^2
  x[[2]] <- a[[3]]*b[[2]]
  sin(x)
}

with gradient (a=c(-1.1,2.3)) {
  x <- numeric(2)
  x[[1]] <- a[[2]]^2
  x[[2]] <- a[[1]]*a[[2]]
  abs(x)
}

with gradient (a=c(1.1,2.3),b=3.4) {
  list (dt(a,5), dt(5,a), dt(a,a), dt(a,b), dt(b,a))
}

with gradient (a=c(1.1,2.3),b=3.4) {
  list (dnorm(a,2,5), dnorm(5,a,2), dnorm(5,2,a), dnorm(a,a,a), 
        dnorm(a,a,b), dnorm(a,b,a), dnorm(b,a,a), dnorm(b,a,b))
}

with gradient (a=c(1.1,2.3)) list (+a, -a)
with gradient (a=c(1.1,2.3)) list (+sin(a), -sin(a))

with gradient (a=c(1.1,2.3),b=c(5.2,3.1)) list (a+b, a-b, a*b, a/b, a^b)
with gradient (a=c(1.1,2.3,3.7),b=c(5.2,3.1)) list (a+b, a-b, a*b, a/b, a^b)
with gradient (a=c(1.1,2.3),b=c(5.2,3.1,1.9)) list (a+b, a-b, a*b, a/b, a^b)
with gradient (a=c(1.1,2.3)) list (1+a, 1-a, a+1, a-1, a+a, a-a)
with gradient (a=c(1.1,2.3)) list (2.1*a, 2.1/a, a*2.1, a/2.1, a*a, a/a)
with gradient (a=c(1.1,2.3)) list (2.1^a, a^2.1, a^a)

with gradient (a=c(1.7,-34.9)) list ((a+2)*(a+1), sin(a)^2+a)
with gradient (a=1.7) list ((a+2)*(a+1), sin(a)^2+a)
with gradient (a=-34.9) list ((a+2)*(a+1), sin(a)^2+a)

with gradient (a=1.7) 
  list (a+3L, a-3L, a*3L, 3L*a, a^3L, a/3L, 3L/a, 3L^a)
with gradient (a=c(1.7,1.7)) 
  list (a+3L, a-3L, a*3L, 3L*a, a^3L, a/3L, 3L/a, 3L^a)
with gradient (a=1.7)
  list (a+3, a-3, a*3, 3*a, a^3, a/3, 3/a, 3^a)
with gradient (a=c(1.7,1.7))
  list (a+3, a-3, a*3, 3*a, a^3, a/3, 3/a, 3^a)

set.seed(179)
with gradient (a=3.1) rexp(1,a)
with gradient (a=0.2) rexp(1,a)
with gradient (a=7.5) rexp(1,a)
with gradient (a=3.1) rexp(1,a^2)
with gradient (a=0.2) rexp(1,a^2)
with gradient (a=7.5) rexp(1,a^2)

set.seed(179)
with gradient (a=c(3.1,0.2,7.5)) rexp(3,a)
with gradient (a=c(3.1,0.2,7.5)) rexp(3,a^2)

set.seed(179)
with gradient (a=3.1,b=2.4) rnorm(1,a,b)
with gradient (a=0.2,b=0.9) rnorm(1,a,b)
with gradient (a=3.1,b=2.4) rnorm(1,cos(a),sin(b)^2)
with gradient (a=0.2,b=0.9) rnorm(1,cos(a),sin(b)^2)
with gradient (a=3.1,b=2.4) rnorm(1,a,3)
with gradient (a=0.2,b=0.9) rnorm(1,a,3)
with gradient (a=3.1,b=2.4) rnorm(1,4,b)
with gradient (a=0.2,b=0.9) rnorm(1,4,b)
with gradient (a=0.7,b=1.8) rnorm(1,a^2,b)
with gradient (a=0.9,b=1.6) rnorm(1,a^2,b)
with gradient (a=0.7,b=1.2) rnorm(1,a^2,b)

set.seed(179)
with gradient (a=c(3.1,0.2),b=c(2.4,0.9)) rnorm(2,a,b)
with gradient (a=c(3.1,0.2),b=c(2.4,0.9)) rnorm(2,cos(a),sin(b)^2)
with gradient (a=c(3.1,0.2),b=c(2.4,0.9)) rnorm(2,a,3)
with gradient (a=c(3.1,0.2),b=c(2.4,0.9)) rnorm(2,4,b)
with gradient (a=c(0.7,0.9),b=c(1.8,1.6,1.2)) rnorm(3,a^2,b)


# Test preservation of gradients through attribute setting.

with gradient (a=c(33,11,0.7,99)) { attr(a,"fred") <- 123; a }
with gradient (a=c(33,11,0.7,99)) { attr(a,"fred") <- NULL; a }

with gradient (a=c(33,11,0.7,99)) { class(a) <- "fred"; a }
with gradient (a=c(33,11,0.7,99)) { class(a) <- NULL; a }

with gradient (a=c(33,11,0.7,99)) { oldClass(a) <- "fred"; a }
with gradient (a=c(33,11,0.7,99)) { oldClass(a) <- NULL; a }

with gradient (a=c(33,11,0.7,99)) { attributes(a) <- list(X=1,y=2); a }
with gradient (a=c(33,11,0.7,99)) { attributes(a) <- NULL; a }

with gradient (a=c(33,11,0.7,99)) { dim(a) <- c(2,2); a }
with gradient (a=c(33,11,0.7,99)) { dim(a) <- NULL; a }

with gradient (a=c(33,11,0.7,99)) { names(a) <- c("ww","xx","yy","zz"); a }
with gradient (a=c(33,11,0.7,99)) { names(a) <- NULL; a }

with gradient (a=matrix(c(33,11,0.7,99),2,2)) {
  dimnames(a) <- list(c("ww","xx"),c("yy","zz"))
  a
}

with gradient (a=matrix(c(33,11,0.7,99),2,2)) { dimnames(a) <- NULL; a }


# Test matrix operations.

(A <- matrix(c(3,6,1,2),2,2))
(B <- matrix(c(9,8,4,3,1,3),2,3))
with gradient (A,B) A %*% B
At <- t(A)
Bt <- t(B)
with gradient (At,B) t(At) %*% B
with gradient (At,B) crossprod(At,B)
with gradient (A,Bt) A %*% t(Bt)
with gradient (A,Bt) tcrossprod(A,Bt)
with gradient (B) t(B) %*% B
with gradient (B) crossprod(B)
with gradient (B) B %*% t(B)
with gradient (B) tcrossprod(B)
(u <- c(11,21))
(v <- c(10,20))
with gradient (A,v) A %*% v
with gradient (A,v) v %*% A
with gradient (u,v) u %*% v
with gradient (u,v) u %*% t(v)
rm(A,B,At,Bt)

with gradient (a=3,b=4) { x <- matrix(c(a,b,a),3,3); drop(x) }
with gradient (a=3,b=4) { x <- matrix(c(a,b,a),9,1); drop(x) }

with gradient (a=matrix(list(3,1,9,4),2,2)) t(a)
with gradient (a=matrix(c(3,1,9,4),2,2)) t(a)
with gradient (a=7) { b <- matrix(c(a,1,a^2,4,a^3,a^4),3,2); t(b) }

with gradient (a=matrix(list(3,1,9,4),2,2)) diag(a)
with gradient (a=matrix(c(3,1,9,4),2,2)) diag(a)

with gradient (a=list(8,7)) diag(a)
with gradient (a=c(8,7)) diag(a)
with gradient (a=8,b=7) diag(c(a,a+3*b,b),3,3)

with gradient (a=3) { M <- matrix(7,2,2); diag(M) <- a; M }
with gradient (a=3) { M <- matrix(7,2,2); diag(M) <- c(a,a^2); M }

with gradient (A=matrix(c(3,1,5,2,-3,9),2,3)) rowSums(A)
with gradient (A=matrix(c(3,1,5,2,-3,9),2,3)) colSums(A)
with gradient (A=matrix(c(3,1,5,2,-3,9),2,3)) rowMeans(A)
with gradient (A=matrix(c(3,1,5,2,-3,9),2,3)) colMeans(A)

with gradient (A=matrix(c(3,1,NA,2,-3,9),2,3)) rowSums(A,na.rm=TRUE)
with gradient (A=matrix(c(3,1,NA,2,-3,9),2,3)) colSums(A,na.rm=TRUE)
with gradient (A=matrix(c(3,1,NA,2,-3,9),2,3)) rowMeans(A,na.rm=TRUE)
with gradient (A=matrix(c(3,1,NA,2,-3,9),2,3)) colMeans(A,na.rm=TRUE)

# next loses dim attribute, which is strange, but accept that for now...
with gradient (a=list(3)) { M <- matrix(7,2,2); diag(M) <- a; M }

X <- matrix(c(3.1,2,7, 0,1.7,5, 9,-3.2,6), 3, 3)
with gradient (X) det(X)
numericDeriv (quote(det(X)), "X")
with gradient (X) determinant(X)
numericDeriv (quote(determinant(X)$modulus), "X")
with gradient (X) determinant(X,logarithm=FALSE)
numericDeriv (quote(determinant(X,logarithm=FALSE)$modulus), "X")

X <- matrix(c(-31,2,7, 0,1.7,5, 9,-3.2,6), 3, 3)
with gradient (X) det(X)
numericDeriv (quote(det(X)), "X")
with gradient (X) determinant(X)
numericDeriv (quote(determinant(X)$modulus), "X")
with gradient (X) determinant(X,logarithm=FALSE)
numericDeriv (quote(determinant(X,logarithm=FALSE)$modulus), "X")

A <- matrix(c(7,1,8.1,2.2),2,2)
v <- c(55,33)
V <- matrix(c(55,33,550,330,5500,3300),2,3)

with gradient (A) solve(A)
numericDeriv (quote(as.vector(solve(A))),"A")

with gradient (A) solve(A,v)
with gradient (v) solve(A,v)
with gradient (A,v) solve(A,v)
numericDeriv (quote(as.vector(solve(A,v))),"A")
numericDeriv (quote(as.vector(solve(A,v))),"v")

with gradient (A) solve(A,V)
with gradient (V) solve(A,V)
with gradient (A,V) solve(A,V)
numericDeriv (quote(as.vector(solve(A,V))),"A")
numericDeriv (quote(as.vector(solve(A,V))),"V")

A <- matrix(c(3,1,7,4),2,2)
B <- matrix(c(4,8,2,9,3,6),2,3)

with gradient (A,B) ((3*A) %*% sin(A+0.1)) %*% (7*B)

numericDeriv (quote(((3*A) %*% sin(A+0.1)) %*% (7*B)), "A")
numericDeriv (quote(((3*A) %*% sin(A+0.1)) %*% (7*B)), "B")



# Tests with data frames.

p <- function (df) { print(df); print(attr(df,"gradient")); }

a <- c(3,7,1)
b <- c(9,2,4)

p( with gradient (a,b) 
     data.frame(x=a,y=c(a[1]-b[1],a[1]+b[1],a[1]*b[2]),z=a+b) )

p( with gradient (a,b) 
     data.frame(list(x=a,y=c(a[1]-b[1],a[1]+b[1],a[1]*b[2]),z=a+b)) )

p( with gradient (a,b) data.frame(matrix(c(a,b,a,a^2),4,3)) )

with gradient (a=c(2,3,1),b=c(9,8,7)) 
  { df <- data.frame(x=a,y=b^2); df$y }
with gradient (a=c(2,3,1),b=c(9,8,7)) 
  { df <- data.frame(x=a,y=b^2); df[,"y"] }
with gradient (a=c(2,3,1),b=c(9,8,7)) 
  { df <- data.frame(x=a,y=b^2); df[2:3,"y"] }
with gradient (a=c(2,3,1),b=c(9,8,7)) 
  { df <- data.frame(x=a,y=b^2); df[[2,"y"]] }

p( with gradient (a=c(2,3,1),b=c(9,8,7)) 
  { df <- data.frame(x=a,y=b^2); df } )
p( with gradient (a=c(2,3,1),b=c(9,8,7)) 
  { df <- data.frame(x=a,y=b^2); df$y <- a^3; df } )
p( with gradient (a=c(2,3,1),b=c(9,8,7)) 
  { df <- data.frame(x=a,y=b^2); df["y"] <- a^3; df } )
p( with gradient (a=c(2,3,1),b=c(9,8,7)) 
  { df <- data.frame(x=a,y=b^2); df[,"y"] <- a^3; df } )
p( with gradient (a=c(2,3,1),b=c(9,8,7)) 
  { df <- data.frame(x=a,y=b^2); df[2:3,"y"] <- a[1..2]^3; df } )
p( with gradient (a=c(2,3,1),b=c(9,8,7)) 
  { df <- data.frame(x=a,y=b^2); df[[2,"y"]] <- a[2]; df } )

df1 <- data.frame(x=a)
df2 <- data.frame(x=b)
df3 <- data.frame(y=b)
p( with gradient (df1,df2) rbind(df1,df2) )
p( with gradient (df1,df3) cbind(df1,df3) )


# Miscellaneous tests.

with gradient (a=7) invisible(a)

with gradient (a=c(3,1,5,4),b=7) mean(c(a,b,a))
with gradient (a=c(3,1,5,4),b=7) mean(c(a,b,NA,a))

with gradient (a=c(3,1,5,4),b=7) cumsum(c(a,b,NA,a))
with gradient (a=c(3,1,5,4),b=7) { x <- a^2; x[2] <- a[3]; cumsum(c(x,b^3)) }
with gradient (a=c(3,1,5,4),b=7) cumprod(c(a,b,NA,a))
with gradient (a=c(3,1,5,4),b=7) { x <- a^2; x[2] <- a[3]; cumprod(c(x,b^3)) }
with gradient (a=c(3,1,5,4),b=7) cummax(c(a,b,NA,a))
with gradient (a=c(3,1,5,4),b=7) { x <- a^2; x[2] <- a[3]; cummax(c(x,b^3)) }
with gradient (a=c(3,1,5,4),b=7) { x <- a^2; x[2] <- sum(a); cummax(c(x,b^3)) }
with gradient (a=c(3,1,5,4),b=7) cummin(c(a,b,NA,a))
with gradient (a=c(3,1,5,4),b=2) { x <- a^2; x[2] <- a[3]; cummin(c(x,b^3)) }
with gradient (a=c(3,1,5,4),b=2) { x <- a^2; x[2] <- -sum(a); cummin(c(x,b^3)) }

with gradient (a=c(3,1,5)) sum(a)
with gradient (a=c(3,1,5)) sum(a^2+a)
with gradient (a=c(3,1,5)) sum(a^2,a)
with gradient (a=c(3,1,5),b=7) sum(a^2+a,3*b)
with gradient (a=c(3,1,NA,5),b=7) sum(a^2+a,3*b,na.rm=FALSE)
with gradient (a=c(3,1,NA,5),b=7) sum(a^2+a,3*b,na.rm=TRUE)

with gradient (a=c(3,1,5)) prod(a)
with gradient (a=c(3,1,5)) prod(a^2+a)
with gradient (a=c(3,1,5)) prod(a^2,a)
with gradient (a=c(3,1,5),b=7) prod(a^2+a,3*b)
with gradient (a=c(3,1,NA,5),b=7) prod(a^2+a,3*b,na.rm=FALSE)
with gradient (a=c(3,1,NA,5),b=7) prod(a^2+a,3*b,na.rm=TRUE)

with gradient (a=c(3,1,5)) max(a)
with gradient (a=c(3,1,5)) max(a-a^2)
with gradient (a=c(3,1,5)) max(a^2,a)
with gradient (a=c(3,1,5),b=7) max(a^2+a,3*b)
with gradient (a=c(3,1,NA,5),b=7) max(a^2+a,3*b,na.rm=FALSE)
with gradient (a=c(3,1,NA,5),b=7) max(a^2+a,3*b,na.rm=TRUE)

with gradient (a=c(3,1,5)) min(a)
with gradient (a=c(3,1,5)) min(a-a^2)
with gradient (a=c(3,1,5)) min(a^2,a)
with gradient (a=c(3,1,5),b=0.1) min(a^2+a,3*b)
with gradient (a=c(3,1,NA,5),b=7) min(a^2+a,3*b,na.rm=FALSE)
with gradient (a=c(3,1,NA,5),b=7) min(a^2+a,3*b,na.rm=TRUE)

a <- c(7,1,8,9)
b <- c(3,4,8,6)

with gradient (a,b) pmin(a,b)
with gradient (a,b) pmax(a,b)
with gradient (a,b) pmax(a,b,a[1]+1)
with gradient (a,b) pmax(a[1]+1,b,a)
with gradient (a,b) pmax(a,b,a*b)

rm(a,b)

d <- c(5,-7,8,-2)

with gradient (d) abs(d+1)
with gradient (d) sin(d+1)
with gradient (d) log(d+10)

rm(d)

with gradient (a=list(3,list(9,2))) as.vector(a)
with gradient (a=c(3,1,9,2)) as.vector(a)

with gradient (a=list(3,2,9)) as.vector(a,"double")
with gradient (a=list(3,2,9)) as.double(a)
with gradient (a=c(3,2,9)) as.vector(a,"list")
with gradient (a=c(3,2,9)) as.list(a)

with gradient (a=list(9,0,2)) { a[[2]] <- "x"; as.vector(a,"double") }

a <- c(9,8); class(a)<-"fred"; as.list.fred <- function(x) list(x[2],x[1])
with gradient (a) as.list(a)
rm(a)

with gradient (a=c(3,9)) { storage.mode(a) <- "double"; a }
with gradient (a=list(3,9)) { storage.mode(a) <- "list"; a }
with gradient (a=c(3,9)) { storage.mode(a) <- "list"; a }
with gradient (a=list(3,9)) { storage.mode(a) <- "double"; a }
with gradient (a=c(3,9)) { storage.mode(a) <- "character"; a }

with gradient (a=list(3,list(9,2),9,8)) matrix(a,2,2)
with gradient (a=list(3,8)) matrix(a,2,2)
with gradient (a=c(3,1,9,2)) matrix(a,2,2)
with gradient (a=12) matrix(a,2,2)
with gradient (a=c(12,10)) matrix(a,3,2)
with gradient (a=12) { M <- matrix(a,2,2); M[1,2] }

with gradient (a=list(3,list(9,2),9,8)) matrix(a,2,2,byrow=TRUE)
with gradient (a=list(3,8)) matrix(a,2,2,byrow=TRUE)
with gradient (a=c(3,1,9,2)) matrix(a,2,2,byrow=TRUE)
with gradient (a=12) matrix(a,2,2,byrow=TRUE)
with gradient (a=c(12,10)) matrix(a,3,2,byrow=TRUE)
with gradient (a=12) { M <- matrix(a,2,2,byrow=TRUE); M[1,2] }

with gradient (a=list(3,list(9,2),9,8)) array(a,c(2,2))
with gradient (a=list(3,8)) array(a,c(2,2))
with gradient (a=c(3,1,9,2,8,1,0,3,9,8,7,6)) array(a,c(2,3,2))
with gradient (a=12) array(a,c(2,2,2))

with gradient (a = list(3,9)) { for (v in a) print(gradient_of(v)); a }
with gradient (a = c(3,1,9)) { for (v in a^2) print(gradient_of(v)); a }

with gradient (a=list(99,33,11)) { length(a) <- 2; a }
with gradient (a=list(99,33,11)) { length(a) <- 3; a }
with gradient (a=list(99,33,11)) { length(a) <- 4; a }

with gradient (a=c(99,33,11)) { length(a) <- 2; a }
with gradient (a=c(99,33,11)) { length(a) <- 3; a }
with gradient (a=c(99,33,11)) { length(a) <- 4; a }

with gradient (a=c(3,5)) rep_len(c(a[2]^2,a[1]^3),length.out=7)
with gradient (a=list(3,5)) rep_len(a,5)

with gradient (a=c(3,5)) rep.int(c(a[2]^2,a[1]^3),c(2,3))
with gradient (a=c(3,5)) rep.int(c(a[2]^2,a[1]^3),2)
with gradient (a=list(3,5)) rep.int(a,c(2,3))
with gradient (a=list(3,5)) rep.int(a,2)

with gradient (a=c(3,5)) rep(c(a[2]^2,a[1]^3),length.out=7)
with gradient (a=c(3,5)) rep(c(a[2]^2,a[1]^3),times=4)
with gradient (a=c(3,5)) rep(c(a[2]^2,a[1]^3),times=c(2,3))
with gradient (a=list(3,5)) rep(c(a[[2]]^2,a[[1]]^3),times=c(2,3))
with gradient (a=list(3,5)) rep(list(a[[2]]^2,a[[1]]^3),times=2)
with gradient (a=list(3,5)) rep(list(a[[2]]^2,a[[1]]^3),times=c(2,3))

rep.fred <- function (x, times) x*times
a <- c(9,10); class(a) <- "fred"
with gradient (a) rep(a,100)
rm(a)

with gradient (a=list(3,5),b=c(9,8),x=2) cbind(a,b,x)
with gradient (a=c(3,5),b=c(9,8),x=2) cbind(a,b,x,a^2)

with gradient (a=list(3,5),b=c(9,8),x=2) rbind(a,b,x)
with gradient (a=c(3,5),b=c(9,8),x=2) rbind(a,b,x,a^2)

with gradient (a=9) { b <- a^2; get_rm(b) }
with gradient (a=9) structure(a^2,fred=9999)

with gradient (a=c(8,9,7)) unlist(a)
with gradient (a=list(8,x=9,7)) unlist(a)
with gradient (a=list(8,x=9,7),b=12) { a[2] <- b^2; unlist(a) }
with gradient (a=list(8,list(9,7))) unlist(a)
with gradient (a=list(8,list(9,7))) unlist(a,recursive=FALSE)

with gradient (a=c(3,8)) c(a,a^2)
with gradient (a=list(3,8)) c(a,a[[2]]^2,a)
with gradient (a=list(3,list(8,7))) c(a,a[[1]]^2)
with gradient (a=list(3,list(8,7))) c(a,a[[1]]^2,recursive=TRUE)

with gradient (a=matrix(c(9,1,4,2),2,2)) apply(a,1,sqrt)
with gradient (a=matrix(c(9,1,4,2),2,2)) apply(a,1,function(x)10*x[1]+x[2])
with gradient (a=matrix(c(9,1,4,2),2,2)) apply(a,2,function(x)list(x,2*x))

with gradient (a=c(36,1,25)) lapply(a,sqrt)
with gradient (a=list(36,1,25)) lapply(a,function(x)c(x,sqrt(x)))
with gradient (a=5) lapply(list(10,100),function(x)x*a)

with gradient (a=c(36,1,25)) vapply(a,sqrt,numeric(1))
with gradient (a=c(36,1,25)) vapply(a,function(x)c(x,sqrt(x)),c(0,0))
with gradient (a=list(36,1,25)) vapply(a,function(x)list(x,sqrt(x)),list(0,0))

with gradient (a=c(36,1,25)) sapply(a,sqrt)
with gradient (a=list(36,1,25)) sapply(a,function(x)c(x,sqrt(x)))

set.seed(1); with gradient (a=2) replicate(10,runif(1,-a,a))

a <- c(4,9)
class(a) <- "fred"
with gradient (a) unclass(a)^2
rm(a)

with gradient (a=c(9,7)) a^2+a[2]^3
with gradient (a=c(9,7)) a[1]^2+a^3

a <- c(3,1,7,4)
b <- a+0.1
with gradient (a) sin(a+0.1)
with gradient (b) sin(b)


# Test backpropagation.

with gradient (a=5) 
  with gradient (d=a^2) sin(d)

with gradient (a=5,b=7) 
  with gradient (d=a^2+b,e=a*b) sin(d)+sqrt(e)

with gradient (a=5,b=7) 
  with gradient (d=a^2+b,e=a*b) list(sin(d)+sqrt(e),55)

with gradient (L=list(a=5,b=7))
  with gradient (d=L$a^2+L$b,e=L$a*L$b) sin(d)+sqrt(e)

with gradient (a=5,b=7)
  with gradient (M=list(d=a^2+b,e=a*b)) sin(M$d)+sqrt(M$e)

with gradient (a=5,b=7)
  with gradient (M=list(d=a^2+b,e=a*b)) list(sin(M$d)+sqrt(M$e),55)

with gradient (L=list(a=5,b=7))
  with gradient (M=list(d=L$a^2+L$b,e=L$a*L$b)) sin(M$d)+sqrt(M$e)

with gradient (L=list(a=5,b=7))
  with gradient (d=L$a^2+L$b,e=L$a*L$b) list(sin(d)+sqrt(e),55)

with gradient (L=list(a=5,b=7))
  with gradient (M=list(d=L$a^2+L$b,e=L$a*L$b)) list(sin(M$d)+sqrt(M$e),55)

with gradient (L=list(a=5,b=7))
  track gradient (M=list(d=L$a^2+L$b,e=L$a*L$b)) list(sin(M$d)+sqrt(M$e),55)

with gradient (L=list(a=5,b=7))
  back gradient (M=list(d=L$a^2+L$b,e=L$a*L$b)) list(sin(M$d)+sqrt(M$e),55)

with gradient (L=c(a=5,b=7)) {
  M <- c(d=0,e=0)
  M[[1]] <- L[["a"]]^2+L[["b"]]
  M[[2]] <- L[["a"]]*L[["b"]]
  back gradient (M) {
    c <- c(0,55)
    c[[1]] <- sin(M[["d"]])+sqrt(M[["e"]])
    c
  }
}

with gradient (a=list(4.1,2.7,3.1,5.2)) {
  back gradient (b=list(a[[2]],sin(a[[1]]+a[[4]]),a[[1]]^2)) {
    c <- list(0,0)
    c[[1]] <- b[[1]] + 4*b[[2]]
    c[[2]] <- b[[3]]^2 + b[[2]]
    c
  }
}

with gradient (a=c(4.1,2.7,3.1,5.2)) {
  b <- numeric(3)
  b[[1]] <- a[[2]]
  b[[2]] <- sin(a[[1]]+a[[4]])
  b[[3]] <- a[[1]]^2
  back gradient (b) {
    c <- numeric(2)
    c[[1]] <- b[[1]] + 4*b[[2]]
    c[[2]] <- b[[3]]^2 + b[[2]]
    c
  }
}


# Test duplication when necessary and not when not.

A <- B <- C <- D <- E <- F <- NULL  # kludge to avoid RPROFMEM messages from
                                    # expansion of hash table for global env

pr <- function (...) { 
  Rprofmemt(NULL)
  cat(...,"\n")
  Rprofmemt(nelem=n,bytes=FALSE)
}

prp <- function (x) { 
  Rprofmemt(NULL)
  print(x)
  Rprofmemt(nelem=n,bytes=FALSE)
}

n <- 100

pr("start")
track gradient (a=rep(3,n)) {
  pr("A"); x <- a+2.1
  pr("B"); x <- a-2.1
  pr("C"); x <- a*2.1
  pr("D"); x <- a/2.1
  pr("E"); x <- a^2.1
  NULL
}
pr("end")

pr("start")
track gradient (b=rep(4,n)) {
  pr("A"); x <- 2.1+b
  pr("B"); x <- 2.1-b
  pr("C"); x <- 2.1*b
  pr("D"); x <- 2.1/b
  pr("E"); x <- 2.1^b
  NULL
}
pr("end")

pr("start")
track gradient (q=7,a=rep(3,n)) {
  pr("A"); x <- a+q
  pr("B"); x <- a-q
  pr("C"); x <- a*q
  pr("D"); x <- a/q
  pr("E"); x <- a^q
  NULL
}
pr("end")

pr("start")
track gradient (q=7,b=rep(4,n)) {
  pr("A"); x <- q+b
  pr("B"); x <- q-b
  pr("C"); x <- q*b
  pr("D"); x <- q/b
  pr("E"); x <- q^b
  NULL
}
pr("end")

pr("start")
track gradient (a=rep(3,n),b=rep(4,n)) {
  pr("A"); x <- a+b
  pr("B"); x <- a-b
  pr("C"); x <- a*b
  pr("D"); x <- a/b
  pr("E"); x <- a^b
  NULL
}
pr("end")

pr("start")
r <- back gradient (a=rep(3,n)) {
  pr("A"); b <- a^2.1
  pr("B"); b[2] <- 7
  c <- b
  pr("C"); b[3] <- 8
  pr("D"); b
}
pr("end")
r[1..5]; attr(r,"gradient")[1..5,1..5]

pr("start")
r <- with gradient (a=rep(3,n)) {
  pr("A"); b <- a^2.1
  pr("B"); b[2] <- 7
  c <- b
  pr("C"); b[3] <- 8
  pr("D"); b
}
pr("end")
r[1..5]; attr(r,"gradient")[1..5,1..5]

pr("start")
r <- with gradient (a=rep(3,n)) {
  pr("A"); L <- list(x=2,y=a^2.1)
  pr("B"); L$y[2] <- 7
  M <- L
  pr("C"); L$y[3] <- 8
  pr("D"); L$x <- a[2]
  pr("E"); L
}
pr("end")
r$x; r$y[1..5]; attr(r,"gradient")$x[1..5]; attr(r,"gradient")$y[1..5,1..5]

pr("start")
r <- with gradient (a=rep(3,n)) {
  pr("A"); L <- list(x=2,y=a^2.1)
  pr("B"); L[[2]][[2]] <- 7
  M <- L
  pr("C"); L[[2]][[3]] <- 8
  pr("D"); L[[1]] <- a[[2]]
  pr("E"); L
}
pr("end")
r$x; r$y[1..5]; attr(r,"gradient")$x[1..5]; attr(r,"gradient")$y[1..5,1..5]

pr("start")
r <- with gradient (a=rep(3,n)) {
  pr("A"); L <- c(list(x=2,y=a^2.1),rep(list(NULL),n))
  pr("B"); L$y[2] <- 7
  M <- L
  pr("C"); L$y[3] <- 8
  pr("D"); L$x <- a[2]
  pr("E"); L
}
pr("end")
r$x; r$y[1..5]; attr(r,"gradient")$x[1..5]; attr(r,"gradient")$y[1..5,1..5]

pr("start")
r <- with gradient (a=rep(3,n)) {
  pr("A"); L <- c(list(x=2,y=a^2.1),rep(list(NULL),n))
  pr("B"); L[[2]][[2]] <- 7
  M <- L
  pr("C"); L[[2]][[3]] <- 8
  pr("D"); L[[1]] <- a[[2]]
  pr("E"); L
}
pr("end")
r$x; r$y[1..5]; attr(r,"gradient")$x[1..5]; attr(r,"gradient")$y[1..5,1..5]

pr("start")
r <- with gradient (a=rep(3,n)+(1..n)) {
  pr("A"); L <- c(list(x=2,y=a^2.1),rep(list(NULL),n))
  pr("B"); q <- list(a[4])
  pr("C"); L[2..3] <- q
  M <- L
  pr("D"); L[2..4] <- list(a[5])
  pr("E"); L[2..3] <- list(a[6])
  pr("F"); L
}
pr("end")
r$x; r$y; 
attr(r,"gradient")$x[1..10]
attr(r,"gradient")$y[1..10]
attr(r,"gradient")[[3]][1..10]
attr(r,"gradient")[[4]][1..10]
attr(r,"gradient")[[5]][1..10]

pr("start")
r <- back gradient (a=rep(3,n)) {
  pr("A"); a[2] <- 7
  pr("B"); b <- c(a,a)
  pr("C"); a[3] <- a[4]^2
  pr("D"); c(a,b)
}
pr("end")
r[1..10]

pr("start")
r <- with gradient (a=rep(3,n)) {
  pr("A"); a[2] <- 7
  pr("B"); b <- c(a,a)
  pr("C"); a[3] <- a[4]^2
  pr("D"); c(a,b)
}
pr("end")
r[1..10]
attr(r,"gradient")[1..10,1..10]

n <- 8

pr("start")
track gradient (a=rep(3,n)) {
  pr("A"); L <- list(a,a,a); prp(gradient_of(L))
  pr("B"); L[[2]][1] <- 9; prp(gradient_of(L))
  pr("C"); M <- L; L[[3]][2] <- L[[1]][3]; prp(gradient_of(L))
  pr("D"); prp(gradient_of(M))
  NULL
}
pr("end")

pr("start")
track gradient (a=rep(3,n)) {
  pr("A"); a[4] <- a[1]+a[2]
  pr("B"); b <- a
  pr("C"); b[2] <- a[5]
  pr("D"); b[7] <- b[8]
  pr("E"); b[1] <- b[6]+a[4]
  pr("F"); a[6] <- a[1]
  pr("G"); x <- a; x <- 7; a[8] <- 9
  pr("H"); prp(gradient_of(a))
  pr("I"); prp(gradient_of(b))
  NULL
}
pr("end")

pr("start")
track gradient (a=3,b=rep(4,n)) {
  prp(x <- gradient_of(b))
  b[2] <- a
  prp(y <- gradient_of(b))
  q <- a*b
  prp(z <- gradient_of(q))
  q[7..8] <- b[1..2]
  prp(w <- gradient_of(q))
  list(x,y,z,w)
}
pr("end")

pr("start")
track gradient (a=7,b=6) {
  pr("A"); L <- list(x=list(y=a,z=a)); prp(gradient_of(L))
  pr("B"); L$x$z <- 2*b; prp(gradient_of(L))
  pr("C"); L[[1]]$z <- 3*b; prp(gradient_of(L))
  pr("D"); L[[1]][[2]] <- 4*b; prp(gradient_of(L))
  pr("E"); L[[1]]$z <- 5*b; prp(gradient_of(L))
  NULL
}
pr("end")

pr("start")
track gradient (a=7,b=6) {
  pr("A"); L <- list(x=list(y=a,z=c(a,b,a,b,a,a,b,b))); prp(gradient_of(L))
  pr("B"); L$x$z[2] <- a+b; prp(gradient_of(L))
  pr("C"); L$x$z[5] <- a-b; prp(gradient_of(L))
  pr("D"); L$x$z[2] <- a*b; prp(gradient_of(L))
  pr("E"); M <- L; L$x$z[3] <- a^2+b^2; prp(gradient_of(L))
  pr("F"); prp(gradient_of(M))
  NULL
}
pr("end")

pr("start")
track gradient (a=7,b=6) {
  pr("A"); L <- list(x=list(y=a,z=c(a,b,a,b,a,a,b,b))); prp(gradient_of(L))
  pr("B"); L[[1]][[2]][[2]] <- a+b; prp(gradient_of(L))
  pr("C"); L[[1]][[2]][[5]] <- a-b; prp(gradient_of(L))
  pr("D"); L[[1]][[2]][[2]] <- a*b; prp(gradient_of(L))
  pr("E"); M <- L; L[[1]][[2]][[3]] <- a^2+b^2; prp(gradient_of(L))
  pr("F"); prp(gradient_of(M))
  NULL
}
pr("end")

pr("start")
track gradient (a=7,b=6) {
  pr("A"); L <- list(x=list(y=a,z=c(a,b,a,b,a,a,b,b)));
  pr("B"); L$x$z[2] <- a+b;
  pr("C"); L$x$z[5] <- a-b;
  pr("D"); L$x$z[2] <- a*b;
  pr("E"); M <- L; L$x$z[3] <- a^2+b^2; prp(gradient_of(L))
  pr("F"); prp(gradient_of(M))
  NULL
}
pr("end")

pr("start")
track gradient (a=7,b=6) {
  pr("A"); L <- list(x=list(y=a,z=c(a,b,a,b,a,a,b,b)));
  pr("B"); L[[1]][[2]][[2]] <- a+b;
  pr("C"); L[[1]][[2]][[5]] <- a-b;
  pr("D"); L[[1]][[2]][[2]] <- a*b;
  pr("E"); M <- L; L[[1]][[2]][[3]] <- a^2+b^2; prp(gradient_of(L))
  pr("F"); prp(gradient_of(M))
  NULL
}
pr("end")

pr("start")
track gradient (a=7,b=6) {
  pr("A"); L <- list(x=list(y=a,z=c(a,b,a,b,a,a,b,b)));
  pr("B"); L$x[[2]][2] <- a+b; # L$x[[2]][2] <- a+b;
  pr("C"); L$x[[2]][5] <- a-b;
  pr("D"); L$x[[2]][2] <- a*b;
  pr("E"); M <- L; L[[1]][[2]][[3]] <- a^2+b^2; prp(gradient_of(L))
  pr("F"); prp(gradient_of(M))
  NULL
}
pr("end")

pr("start")
track gradient (a=7,b=6) {
  pr("A"); L <- list(x=list(y=a,z=c(a,b,a,b,a,a,b,b)))
  pr("B"); L$x$z[[2]] <- a+b
  pr("C"); L$x$z[[5]] <- a-b
  pr("D"); L$x$z[[2]] <- a*b
  pr("E"); M <- L; L$x$z[3] <- a^2+b^2; prp(gradient_of(L))
  pr("F"); prp(gradient_of(M))
  NULL
}
pr("end")

pr("start")
track gradient (a=rep(7,n)) {
  cat("W\n")
  r <- compute gradient (a) {
    cat("X\n")
    list(5,1,9)
  }
  as {
    cat("Y\n")
    list(a*2,a*3,a*4)
  }
  cat("Z\n")
  NULL
}
pr("end")

pr("start")
with gradient (a=c(3,9,7,4,2,6,5,8)) {
  pr("A"); L <- list(x=list(y=a))
  pr("B"); b <- a
  pr("C"); b[2] <- b[1]  # makes jacobian for b non-diagonal
  pr("D"); L$x$y <- a+1
  pr("E"); L$x$y <- 3*a
  pr("F"); L$x$y <- b
  L
}
pr("end")

pr("start")
with gradient (L=list(x=list(y=c(3,9,7,4,2,6,5,8)))) {
  pr("A"); b <- L$x$y
  pr("B"); b[2] <- b[1]  # makes jacobian for b non-diagonal
  pr("F"); L$x$y <- b
  L
}
pr("end")

pr("start")
with gradient (a=c(3,9,7,4,2,6,5,8)+0.1) {
  pr("A"); x <- a^a
  pr("B"); y <- x^(1/a)
  pr("C"); z <- dnorm(a,a+1,a+2)
  pr("D"); list(x,y,z)
}
pr("end")

pr("start")
with gradient (a=c(3,9,0.04,4,0.16,6,5,8)) {
  pr("A"); b <- c(a[2],a[-2])  # has non-diagonal jacobian
  pr("B"); w <- pmax(a,sqrt(a))
  pr("C"); x <- pmax(b,b)
  pr("D"); y <- pmax(a,b)
  pr("E"); z <- pmax(b,a)
  pr("F"); list(w,x,y,z)
}
pr("end")

pr("start")
with gradient (a=c(3,9,7,4,2,6,5,8)+0.1) {
  pr("A"); x <- mean(a)
  pr("B"); y <- mean(a^2)
  pr("C"); b <- a; b[2] <- b[8]
  pr("D"); z <- mean(b)
  pr("E"); list(x,y,z)
}
pr("end")

pr("start")
with gradient (a=c(3,9,7,4,2,6,5,8)+0.1) {
  pr("A"); w <- 1-a
  pr("B"); x <- 10-w
  pr("C"); y <- cumsum(a)
  pr("D"); z <- -y
  pr("E"); list(x,y,z)
}
pr("end")

pr("start")
with gradient (a=c(3,9,7,4,2,6,5,8)) {
  pr("A"); b <- a^2
  pr("B"); w <- 2*b
  pr("C"); x <- 5*w
  pr("D"); y <- cumsum(b)
  pr("E"); z <- 100*y
  pr("F"); list(x,y,z)
}
pr("end")

pr("start")
with gradient (a=c(3,9,7,4,2,6,5,8),
               b=c(1,10,100,1000,10000,100000,1000000,10000000)) {
  pr("A"); x <- 2*a
  pr("B"); y <- a*b
  pr("C"); z <- a/b
  pr("D"); q <- c(a[1]*a[3],a[3],a[2],a[2]*a[8],a[1]*a[3],a[4],a[5],a[7]*a[8])
  pr("E"); w <- q*b
  pr("F"); list(x,y,z,q,w)
}
pr("end")

pr("start")
with gradient (a=c(3,9,7,4,2,6,5,8),b=c(3,9,7,4,2,6,5,8)) {
  pr("A"); x <- cumsum(a)
  pr("B"); y <- (c(10,20,30,40,50,60,70,80)*x) * c(8,7,6,5,4,3,2,1)
  pr("C"); z <- a*x
  pr("D"); w <- b*x
  pr("E"); list(x,y,z,w)
}
pr("end")

pr("start")
with gradient (a=c(3,9,7,4,2,6,5,8)) {
  pr("A"); x <- cumsum(a)
  pr("B"); y <- x^2 + sin(x)
  pr("C"); z <- sqrt(dnorm(x,0,x+1))
  pr("D"); list(x,y,z)
}
pr("end")

pr("start")
with gradient (a=c(3,9,7,4,2,6,5,8)) {
  pr("A"); w <- cumsum(a)
  pr("B"); x <- 7*w
  pr("C"); y <- cumsum(x)
  pr("D"); z <- 3*x
  pr("E"); list(w,x,y,z)
}
pr("end")

pr("start")
with gradient (a=c(3,9,7,4,2,6,5,8)) {
  pr("A"); x <- cumsum(a)
  pr("B"); y <- 3*x
  pr("C"); z <- x + y
  pr("D"); z <- z + x
  pr("E"); list(x,y,z)
}
pr("end")

pr("start")
with gradient (a=c(3,9,7,4,2,6,5,8)) {
  pr("A"); x <- cumsum(a)
  pr("B"); y <- x^2
  pr("C"); z <- y + x
  pr("D"); z <- x + z
  pr("E"); list(x,y,z)
}
pr("end")

pr("start")
with gradient (a=c(3,9,7,4,2,6,5,8)) {
  pr("A")
  r <- back gradient (b = a^2) {
    pr("B"); 3*b
  }
  pr("C"); r
}
pr("end")

pr("start")
with gradient (a=c(3,9,7,4,2,6,5,8)) {
  pr("A")
  r <- back gradient (b = a^2, c = a^3) {
    pr("B"); 3*b + 4*c
  }
  pr("C"); r
}
pr("end")

pr("start")
with gradient (a=c(3,9,7,4,2,6,5,8)) {
  pr("A"); b <- a^2
  pr("B"); r <- 3*b
  pr("C"); r
}
pr("end")

pr("start")
with gradient (a=c(3,9,7,4,2,6,5,8)) {
  pr("A"); x <- a[3]
  pr("B"); r <- a*x
  pr("C"); s <- x*a
  pr("D"); list(r,s)
}
pr("end")

pr("start")
with gradient (a=c(3,9,7,4,2,6,5,8)) {
  pr("A")
  r <- compute gradient (b = a^2)
         { pr("B"); 3*b }
       as { pr("C"); diag(rep(3,length(b))) }
  pr("D"); r
}
pr("end")

pr("start")
with gradient (a=c(3,9,7,4,2,6,5,8)) {
  pr("A")
  r <- compute gradient (b = a^2)
         { pr("B"); 3*b }
       as { pr("C"); function () diag(rep(3,length(b))) }
  pr("D"); r
}
pr("end")

pr("start")
with gradient (a=c(3,9,7,4,2,6,5,8)) {
  pr("A")
  r <- compute gradient (b = a^2)
         { pr("B"); 3*b }
       as { pr("C"); rep(3,length(b)) }
  pr("D"); r
}
pr("end")

pr("start")
with gradient (a=c(3,9,7,4,2,6,5,8)) {
  pr("A")
  r <- compute gradient (b = a^2)
         { pr("B"); 3*b }
       as { pr("C"); function () rep(3,length(b)) }
  pr("D"); r
}
pr("end")

pr("start")
with gradient (a=c(3,9,7,4,2,6,5,8)) {
  pr("A"); b <- a[3..10]
  pr("B"); c <- b[2..4]
  pr("C"); d <- 3*c
  pr("D"); list(b,c,d)
}
pr("end")

pr("start")
with gradient (a=c(3,9,7,4,2,6,5,8)) {
  pr("A"); b <- 7*cumsum(a)[3..10]
  pr("B"); c <- b[2..5]
  pr("C"); d <- c[2..3]
  pr("D"); e <- 3*d
  pr("E"); list(b,c,d,e)
}
pr("end")

pr("start")
with gradient (a=c(3,9,7,4,2,6,5,8)) {
  pr("A"); b <- a[c(3,4,5,6,7,8,9,10)]
  pr("B"); c <- b[c(2,3,4)]
  pr("C"); d <- 3*c
  pr("D"); list(b,c,d)
}
pr("end")

pr("start")
with gradient (a=c(3,9,7,4,2,6,5,8)) {
  pr("A"); b <- 7*cumsum(a)[c(3,4,5,6,7,8,9,10)]
  pr("B"); c <- b[c(2,3,4,5)]
  pr("C"); d <- c[c(2,3)]
  pr("D"); e <- 3*d
  pr("E"); list(b,c,d,e)
}
pr("end")

pr("start")
with gradient (a=c(3,9,7,4,2,6,5,8)) {
  pr("A"); b <- a[3]
  pr("B"); c <- b + 2*a[3]
  pr("C"); d <- b + 2*a[4]
  pr("D"); list(b,c,d)
}
pr("end")

pr("start")
with gradient (a=c(3,9,7,4,2,6,5,8)) {
  pr("A"); b <- numeric(10);
  pr("B"); b[[3]] <- a[[5]];
  pr("C"); b[[7]] <- a[[2]];
  pr("D"); c <- 3*b
  pr("E"); list(b,c)
}
pr("end")

pr("start")
with gradient (a=c(3,9,7,4,2,6,5,8)) {
  pr("A"); b <- numeric(10);
  pr("B"); b[[3]] <- a[5];
  pr("C"); b[[7]] <- a[2];
  pr("D"); c <- 3*b
  pr("E"); list(b,c)
}
pr("end")

pr("start")
with gradient (a=c(3,9,7,4,2,6,5,8)) {
  pr("A"); b <- numeric(10);
  pr("B"); b[3..4] <- a[5];
  pr("C"); b[7..9] <- a[2..4];
  pr("D"); c <- 3*b
  pr("E"); list(b,c)
}
pr("end")

pr("start")
with gradient (a=c(3,9,7,4,2,6,5,8)) {
  pr("A"); b <- numeric(10);
  pr("B"); b[c(4,3)] <- a[5];
  pr("C"); b[c(7,9,8)] <- a[c(2,4,3)];
  pr("D"); c <- 3*b
  pr("E"); list(b,c)
}
pr("end")

pr("start")
with gradient (a=c(3,9,7,4,2,6,5,8)) {
  pr("A"); b <- numeric(10);
  pr("B"); b[c(4,3)] <- a[5]+a[7];
  pr("C"); b[c(7,9,8)] <- a[c(2,4,3)];
  pr("D"); c <- 3*b
  pr("E"); list(b,c)
}
pr("end")

pr("start")
with gradient (a=c(3,9,7,4,2,6,5,8)) {
  pr("A"); b <- a[1]+a[8]
  pr("B"); r <- back gradient (b) rep(b,9)
  pr("C"); r
}
pr("end")

# Chains that might be done in reverse mode.

(A <- matrix(c(3,6,1,2),2,2))
(B <- matrix(c(9,8,4,3,1,3),2,3))
pr("start")
with gradient (A,B) {
  pr("A"); X <- A %*% B
  pr("B"); Y <- A %*% X
  pr("C"); list(X,Y)
}
pr("end")

(A <- matrix(c(3,6,1,2),2,2))
(B <- matrix(c(9,8,4,3,1,3),2,3))
(C <- matrix(c(4,3,7),3,1))
pr("start")
with gradient (A,B,C) {
  pr("A"); X <- A %*% B
  pr("B"); Y <- X %*% C
  pr("C"); Y
}
pr("end")

(A <- matrix(c(3,6,1,2),2,2))
(B <- matrix(c(9,8,4,3,1,3),2,3))
(C <- matrix(c(4,3,7),3,1))
pr("start")
with gradient (A,B,C) {
  pr("A"); X <- B %*% C
  pr("B"); Y <- A %*% X
  pr("C"); Y
}
pr("end")

(A <- matrix(c(4,3,7),1,3))
(B <- matrix(c(9,8,4,3,1,3),3,2))
(C <- matrix(c(3,6,1,2),2,2))
pr("start")
with gradient (A,B,C) {
  pr("A"); X <- A %*% B
  pr("B"); Y <- X %*% C
  pr("C"); Y
}
pr("end")

(A <- matrix(c(4,3,7),1,3))
(B <- matrix(c(9,8,4,3,1,3),3,2))
(C <- matrix(c(3,6,1,2),2,2))
pr("start")
with gradient (A,B,C) {
  pr("A"); X <- B %*% C
  pr("B"); Y <- A %*% X
  pr("C"); Y
}
pr("end")

(A <- matrix(c(4,3,7),1,3))
(B <- matrix(c(9,8,4,3,1,3),3,2))
(C <- matrix(c(3,6,1,2),2,2))
(D <- matrix(c(7,3,4,6,1,2,-2,8),2,4))
pr("start")
with gradient (D) {
  pr("A"); X <- C %*% D
  pr("B"); Y <- B %*% X
  pr("C"); Z <- A %*% Y
  pr("D"); Z
}
pr("end")
pr("start")
with gradient (A,B,C,D) {
  pr("A"); X <- C %*% D
  pr("B"); Y <- B %*% X
  pr("C"); Z <- A %*% Y
  pr("D"); Z
}
pr("end")

(A <- matrix(c(4,3,7),1,3))
(B <- matrix(c(9,8,4,3,1,3),3,2))
(C <- matrix(c(3,6,1,2),2,2))
(D <- matrix(c(7,3,4,6,1,2,-2,8),2,4))
pr("start")
with gradient (A) {
  pr("A"); X <- A %*% B
  pr("B"); Y <- X %*% C
  pr("C"); Z <- Y %*% D
  pr("D"); Z
}
pr("end")
pr("start")
with gradient (A,B,C,D) {
  pr("A"); X <- A %*% B
  pr("B"); Y <- X %*% C
  pr("C"); Z <- Y %*% D
  pr("D"); Z
}
pr("end")

pr("start")
with gradient (D = matrix(c(7,3,4,6,1,2,-2,8),2,4)) {
  A <- D[1..1,2..4]
  B <- D[c(1,2,1),3..4]
  C <- D[_,c(4,2)]
  pr("A"); X <- C %*% D
  pr("B"); Y <- B %*% X
  pr("C"); Z <- A %*% Y
  pr("D"); Z
}
pr("end")

A <- matrix(c(3,6,8,2),2,2)
B <- A+1
Bt <- t(B)

pr("start")
with gradient (A) B %*% (B %*% A)
with gradient (A) B %*% crossprod(Bt,A)
with gradient (A) crossprod (Bt, B %*% A)
with gradient (A) crossprod (Bt, crossprod (Bt,A))
with gradient (A) (A %*% B) %*% B
with gradient (A) tcrossprod(A,Bt) %*% B
with gradient (A) tcrossprod (A %*% B, Bt)
with gradient (A) tcrossprod (tcrossprod(A,Bt), Bt)
pr("end")

pr("start")
with gradient (A) B %*% (3 * (B %*% A))
with gradient (A) B %*% ((B %*% A) * 3)
with gradient (A) (3 * (B %*% A)) %*% B
with gradient (A) ((B %*% A) * 3) %*% B
pr("end")

A <- matrix(33,10,10)
A2 <- matrix(22,10,10)
B <- matrix(47,2,10)
Bt <- t(B)

pr("start")
with gradient (A) B %*% (3 * A)
with gradient (A) (3 * A) %*% Bt
with gradient (A) B %*% (3 * (A2 %*% A))
with gradient (A) (3 * (A2 %*% A)) %*% Bt
pr("end")

Rprofmemt(NULL)

x <- rep(c(0.08,-0.04,0.14),5)
W <- matrix(c(4,2),5,15)
Wt <- t(W)
b <- c(0.4,0.3,0.1,-0.3,0.2)
u <- c(2,-2,-1,4,5)

numericDeriv (quote (u %*% tanh (b + W %*% x)), "W")
numericDeriv (quote (u %*% tanh (b + W %*% x)), "b")
numericDeriv (quote (u %*% tanh (b + W %*% x)), "u")

pr("start")
with gradient (W,b,u) u %*% tanh (b + W %*% x)
pr("end")

pr("start")
with gradient (W,b,u) sum (u * tanh (b + W %*% x))
pr("end")

pr("start")
with gradient (Wt,b,u) tanh (b + x %*% Wt) %*% u
pr("end")

pr("start")
with gradient (Wt,b,u) sum (u * tanh (b + x %*% Wt))
pr("end")

A <- matrix(c(3,1,7,2),2,2)
B <- matrix(c(10,20,30,-10),2,2)

with gradient (A) { C <- B%*%A; dim(C) <- c(1,4); c(100,200,500,700)%*%C }
with gradient (A) { C <- A%*%B; dim(C) <- NULL; C%*%c(100,200,500,700) }

Rprofmemt(NULL)

n <- 4

ff <- function (x) compute gradient (x) c(3,4,2) as rep(c(7,2,8),length(x))
gg <- function (x) compute gradient (x) sum(x) as rep(1,length(x))

pr("start")
with gradient (x=c(3,1,4,2)) gg(ff(ff(ff(x))))
pr("end")

pr("start")
with gradient (x=c(3,1,4,2)) sum(ff(ff(ff(x))))
pr("end")

pr("start")
with gradient (x=c(3,1,4,2)) c(1,1,1) %*% ff(ff(ff(x)))
pr("end")

pr("start")
with gradient (x=c(3,1,4,2)) ff(ff(ff(x))) %*% c(1,1,1)
pr("end")

Rprofmemt(NULL)

n <- 100

F <- seq(from=1,by=0.01,length=n)
F1 <- rep(1,length=n)

f <- function (x) 2*x

pr("start")
r1 <- with gradient (a = numeric(n)) f(f(a))
print(attr(r1,"gradient")[1..3,1..3])
pr("end")

pr("start")
r1 <- with gradient (a = numeric(n)) f(f(f(a)))
print(attr(r1,"gradient")[1..3,1..3])
pr("end")

pr("start")
r1 <- with gradient (a = numeric(n)) F*f(10*f(F*f(a)))
print(attr(r1,"gradient")[1..3,1..3])
pr("end")

g <- function (x) compute gradient (x) 2*x as D

D <- diag(2,n)

pr("start")
r2 <- with gradient (a = numeric(n)) g(g(a))
print(attr(r2,"gradient")[1..3,1..3])
pr("end")

pr("start")
r2 <- with gradient (a = numeric(n)) g(g(g(a)))
print(attr(r2,"gradient")[1..3,1..3])
pr("end")

pr("start")
r2 <- with gradient (a = numeric(n)) F*g(10*g(F*g(a)))
print(attr(r2,"gradient")[1..3,1..3])
pr("end")

D <- rep(2,n)

pr("start")
r2 <- with gradient (a = numeric(n)) g(g(a))
print(attr(r2,"gradient")[1..3,1..3])
pr("end")

pr("start")
r2 <- with gradient (a = numeric(n)) g(g(g(a)))
print(attr(r2,"gradient")[1..3,1..3])
pr("end")

pr("start")
r2 <- with gradient (a = numeric(n)) F*g(10*g(F*g(a)))
print(attr(r2,"gradient")[1..3,1..3])
pr("end")


Rprofmemt(NULL)

n <- 3

Z <- matrix(c(-3,1,3,8,2),6,5)
A <- matrix(c(3,9,7,2,2,-1,3,-2),4,6)
B <- matrix(c(8,4),3,4)
C <- matrix(c(-1,2,5),6,3)
D <- matrix(c(7,5,3,1),8,6)
x <- c(7,-1,2,4,9)

(r <- numericDeriv (quote ({
  Y <- Z %*% x
  Y <- A %*% Y
  Y <- B %*% Y
  Y <- C %*% Y
  Y <- D %*% Y
  Y
}), "x"))
g <- attr(r,"gradient")

pr("start")
(r <- with gradient (x) {
  pr("Z"); Y <- Z %*% x
  pr("A"); Y <- A %*% Y
  pr("B"); Y <- B %*% Y
  pr("C"); Y <- C %*% Y
  pr("D"); Y <- D %*% Y
  pr("E"); Y
})
pr("end")

identical(attr(r,"gradient"),g)

pr("start")
(r <- with gradient (x) {
  back gradient (Y = Z %*% x)
    back gradient (Y = A %*% Y)
      back gradient (Y = B %*% Y)
        back gradient (Y = C %*% Y)
          D %*% Y
})
pr("end")

identical(attr(r,"gradient"),g)
